//=========================================================== A2KArea
//
// Umrechnung von Geruchsstundenhäufigkeiten auf ein
// reguläres Netz von Beurteilungsflächen.
//
// ===============================================================
//
// Copyright (C) Nds. Landesamt für Ökologie, Hildesheim, Germany, 2004
// Copyright (C) Landesumweltamt NRW, Essen, Germany, 2004
// Copyright (C) Landesanstalt für Umweltschutz, Karlsruhe, Germany, 2004
// Copyright (C) Janicke Consulting, 88662 Überlingen, Germany, 2004-2015
// email: info@austal2000g.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//-------------------------------------------------------------------
//
// last change 2016-06-03 uj
//
//===================================================================

package de.janicke.tools;

import java.io.File;
import java.io.FilenameFilter;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Vector;

/**
 * @author IBJ/uj
 * @version 1.3.6 2016-06-03
 *  
 */
public class A2KArea {

  protected static boolean DEBUG = false;
  protected static String version = "1.3.6";
  protected boolean bedge;
  protected String sWork;
  protected File dirWork;
  protected float x0, y0, dd, rr;
  protected int nx, ny;
  protected float xmin, ymin;
  protected AValue[] lstAValue;
  protected A2KArray aaFrq, aaSct;
  protected Hashtable<String,String[]> table;
  protected double[] dg;
  protected double[] xgmin;
  protected double[] xgmax;
  protected double[] ygmin;
  protected double[] ygmax;
  private String sName;

  public A2KArea(String work) {
    this(work, null, null, false);
  }
    
  public A2KArea(String work, String type, String name, boolean edge) {
    sWork = work;
    setGrid(0, 0, 250, 600);
    dirWork = new File(sWork);
    if (!dirWork.isDirectory() || !dirWork.canRead()) {
      dirWork = null;
    }
    if (name == null)                                              //-2014-06-22
      sName = (type == null) ? "odor" : "odor_" + type;
    else
      sName = name;
    bedge = edge;                                                  //-2014-06-22
  }
  
  

  public void setGrid(float x0, float y0, float dd, float rr) {
    this.x0 = x0;
    this.y0 = y0;
    this.dd = dd;
    this.rr = rr;
    nx = (int) Math.ceil(2 * rr / dd);
    ny = nx;
    xmin = x0 - nx * dd / 2;
    ymin = y0 - ny * dd / 2;
  }

  public void setGrid(float xmin, float ymin, float dd, int nx, int ny) {
    this.xmin = xmin;
    this.ymin = ymin;
    this.dd = dd;
    this.nx = nx;
    this.ny = ny;
    rr = Math.max(nx * dd / 2, ny * dd / 2);
  }

  public File getWork() {
    return dirWork;
  }

  //------------------------------------------------------------ makeAList
  public String makeAList() {
    String msg = null, fn = null;
    if (dirWork == null) return "*** no work directory";
    Vector<AValue> vv = new Vector<AValue>();
    String[] flist = dirWork.list(new FilenameFilter() {
      public boolean accept(File dir, String name) {
        return name.matches(sName + "-j00z[0-9]*\\.dmna");
      }
    });
    int nn = flist.length;
    if (nn == 0) 
      return "*** no odor files";
    String d2 = "%02d";
    HashSet<String> set = new HashSet<String>();
    for (int n = 0; n < nn; n++)
      set.add(flist[n]);

    boolean nested = !set.contains(sName + "-j00z.dmna");
    if (!nested && nn > 1)
      return "*** mixed results for simple and nested grids!";
    for (int n = 0; n < nn; n++) {
      String xtn = ((nested) ? String.format(d2, n + 1) : "") + ".dmna";
      fn = sName + "-j00z" + xtn;
      if (!set.contains(fn)) 
        return "*** file \"" + fn + "\" not found";
    }
    dg = new double[nn];
    xgmin = new double[nn];
    xgmax = new double[nn];
    ygmin = new double[nn];
    ygmax = new double[nn];
    try {
      int nt;
      int skip = (bedge) ? 2 : 0;                                  //-2014-06-22
      for (int n = 0; n < nn; n++) {
        nt = 0;
        String xtn = ((nested) ? String.format(d2, n + 1) : "") + ".dmna";
        fn = sName + "-j00z" + xtn;
        A2KArray ah = new A2KArray(dirWork.getPath(), fn);
        if (!ah.valid) 
          return "*** invalid: " + fn;
        if (table == null) table = new Hashtable<String,String[]>(ah.getTable());
        if (ah.getDims() != 3) 
          return "*** invalid structure: " + fn;
        String gn = sName + "-j00s" + xtn;
        A2KArray as = null;
        if (new File(dirWork, gn).exists()) {
          as = new A2KArray(dirWork.getPath(), gn);
          if (!as.valid) 
            return "*** invalid: " + gn;
          if (as.getDims() != 3) return "*** invalid structure: " + gn;
        }
        A2KArray ab = null;
        gn = "volout" + xtn;
        if (new File(dirWork, gn).exists()) {
          ab = new A2KArray(dirWork.getPath(), gn);
          if (!ab.valid) 
            return "*** invalid: " + gn;
          if (ab.getDims() != 3) return "*** invalid structure: " + gn;
        }
        int[] hghb = ah.getUpperBounds();
        int nnx = hghb[0];
        int nny = hghb[1];
        double ddd = ah.getDouble("delt|delta");
        double dh = ddd / 2;
        dg[n] = ddd;
        xgmin[n] = ah.getDouble("xmin");
        xgmax[n] = xgmin[n] + nnx * ddd;
        ygmin[n] = ah.getDouble("ymin");
        ygmax[n] = ygmin[n] + nny * ddd;
        if (DEBUG)
          System.out.println("n=" + n + ", xmin=" + xgmin[n] + ", ymin="
              + ygmin[n] + ", dd=" + ddd + ", nx=" + nnx + ", nny=" + nny);
        if (skip != 0) {       
          xgmin[n] = ah.getDouble("xmin") + skip*ddd;
          xgmax[n] = xgmin[n] + nnx * ddd - 2*skip*ddd;
          ygmin[n] = ah.getDouble("ymin") + skip*ddd;
          ygmax[n] = ygmin[n] + nny * ddd - 2*skip*ddd;
          if (DEBUG)
            System.out.println("n=" + n + ", xmin=" + xgmin[n] + ", ymin="
                + ygmin[n]);
        }
        for (int i = 1+skip; i <= nnx-skip; i++) {
          double x = xgmin[n] + (i - 1 - skip) * ddd;
          double xh = x + dh;
          for (int j = 1+skip; j <= nny-skip; j++) {
            double y = ygmin[n] + (j - 1 - skip) * ddd;
            double yh = y + dh;
            if ((n > 0) && (xh > xgmin[n - 1]) && (xh < xgmax[n - 1])
                && (yh > ygmin[n - 1]) && (yh < ygmax[n - 1])) continue;
            double vh = ah.value(i, j, 1);
            double sh = (as != null) ? as.value(i, j, 1) : 0;
            if (ab != null) {
              if (ab.value(i, j, 1) > 0) {
                vh = -1; 
                sh = -1;
              }
            }
            AValue av = new AValue(x, y, ddd, vh, sh);
            vv.add(av);
            nt++;
          }
        }
        if (DEBUG) System.out.println(nt + " tiles from grid " + (n + 1));
      }
      lstAValue = vv.toArray(new AValue[]{});
    } catch (Exception e) {
      e.printStackTrace();
      msg = e.getMessage();
    }
    return msg;
  }

  public AValue[] getAList() {
    return lstAValue;
  }

  //----------------------------------------------------------- makeArray
  public String makeArray() {
    String msg = null;
    if (lstAValue == null) msg = makeAList();
    if (msg != null) return msg;
    try {
      int nn = lstAValue.length;
      if (DEBUG) System.out.println(nn + " entries");
      aaFrq = new A2KArray(1, nx, 1, ny);
      aaSct = new A2KArray(1, nx, 1, ny);
      for (int i = 1; i <= nx; i++) {
        double x1 = xmin + i * dd;
        double xx0 = x1 - dd;
        for (int j = 1; j <= ny; j++) {
          double y1 = ymin + j * dd;
          double yy0 = y1 - dd;
          double h = 0;
          double s = 0;
          double w = 0;
          double u = 0;
          for (int n = 0; n < nn; n++) {
            AValue av = lstAValue[n];
            double dx = getOverlap(xx0, x1, av.x, av.x + av.d);
            if (dx <= 0) continue;
            double dy = getOverlap(yy0, y1, av.y, av.y + av.d);
            if (dy <= 0) continue;
            double a = dx * dy;
            if (av.h < 0) {
              u += a;
            }
            else {
              h += a * av.h;
              s += a * av.s * av.s;
              w += a;
            }
          }
          if ((w + u)/(dd*dd) < 0.99) {
            if (DEBUG)
              System.out.println("i=" + i + ", j=" + j + ", w=" + w + ", dd="
                  + dd);
            return "*** no complete overlap";
          }
          if ((w + u)/(dd*dd) > 1.01) return "*** overlap error";
          if (w > 0) {
            h /= w;
            s /= w;
          }
          aaFrq.setValue(i, j, (float)h);
          aaSct.setValue(i, j, (float)Math.sqrt(s));
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
      msg = e.getMessage();
    }
    return msg;
  }

  //---------------------------------------------------------------- save
  
  private void putl(String key, double value) {
    String[] ss = (String[])table.get("locl");
    String locl = (ss == null) ? "C" : ss[0];
    String v = "" + value;
    if (locl.equalsIgnoreCase("german") || 
        locl.equalsIgnoreCase("\"german\"") ||
        locl.equalsIgnoreCase("de") ||                           //-2012-09-04
        locl.equalsIgnoreCase("\"de\""))
      v = v.replace('.', ',');
    table.put(key, new String[] { v });
  }
  
  public String save(String file, boolean verbose) {
    String msg = null;
    try {
      table.remove("dd");
      table.remove("delt");
      putl("delta", dd);
      putl("xmin", xmin);
      putl("ymin", ymin);
      aaFrq.setTable(table);
      aaSct.setTable(table);
      aaFrq.save(file + "z.dmna");
      if (verbose) System.out.println(file + "z.dmna saved");
      aaSct.save(file + "s.dmna");                            //-2004-09-06
      if (verbose) System.out.println(file + "s.dmna saved");
    } catch (Exception e) {
      e.printStackTrace();
      msg = "*** error saving " + file;
    }
    return msg;
  }

  public static double getOverlap(double x0, double x1, double a, double b) {
    double d = 0;
    if (a > x1 || b < x0) return 0;
    if (a < x0) {
      if (b > x1)
        d = x1 - x0;
      else
        d = b - x0;
    } else {
      if (b > x1)
        d = x1 - a;
      else
        d = b - a;
    }
    return d;
  }

  //=============================================================== AValue
  public static class AValue {

    double x, y, d, h, s;

    public AValue(double x, double y, double d, double h, double s) {
      this.x = x;
      this.y = y;
      this.d = d;
      this.h = h;
      this.s = s;
    }
  }

  //=======================================================================

  public static void help() {
    System.out.print("A2KArea " + version
        + " - Umrechnung von Geruchsstundenhaeufigkeiten\n"
        + "auf ein regulaeres Netz von Beurteilungsflaechen.\n\n"
        + "Aufruf: java -jar A2KArea.jar <Verzeichnis> <Optionen ...>\n"
        + "Verzeichnis: Arbeitsverzeichnis von AUSTAL2000 mit odor-Dateien\n"
        + "Optionen:    -d<D> Maschenweite des neuen Netzes\n"
        + "             -e    Zwei aeusserste Maschen nicht beruecksichtigen\n"
        + "             -N<Name> Expliziter Substanzname\n\n"
        + "Bei Vorgabe eines kreisfoermigen Auswertegebietes:\n"
        + "             -r<R> Ausdehnung (Radius) des neuen Netzes\n"
        + "             -x<X> x-Koordinate des Netz-Mittelpunktes\n"
        + "             -y<Y> y-Koordinate des Netz-Mittelpunktes\n\n"
        + "Bei Vorgabe eines rechteckigen Auswertegebietes:\n"
        + "             -n<Nx>,<Ny> Anzahl der Intervalle in x- und y-Richtung\n"
        + "             -x<X> x-Koordinate des linken Netz-Randes\n"
        + "             -y<Y> y-Koordinate des unteren Netz-Randes\n\n"
        + "optional:    -t<type>  Kennzeichen der Geruchsart\n\n"
        + "Die Namen der Ergebnis-Dateien sind:\n"
        + "\"odor[_<type>]-d<D>r<R>x<X>y<Y>-z.dmna\" fuer die Geruchsstundenhaeufigkeit\n"
        + "\"odor[_<type>]-d<D>r<R>x<X>y<Y>-s.dmna\" fuer deren Streuung.\n\n"
        + "beziehungsweise bei einem rechteckigen Auswertegebiet:\n"
        + "\"odor[_<type>]-d<D>x<X>n<Nx>y<Y>n<Ny>-z.dmna\" fuer die Geruchsstundenhaeufigkeit\n"
        + "\"odor[_<type>]-d<D>x<X>n<Nx>y<Y>n<Ny>-s.dmna\" fuer deren Streuung.\n\n"
       );
    System.exit(0);
  }

  public static void main(String[] args) {
    if (args.length == 0) {
      help();
    }
    float x = 0, y = 0, d = 250, r = 600;
    int nx = 0, ny = 0;
    String work = null;
    String msg = null, s = null;
    String type = null;
    String name = null;
    boolean edge = false;
    try {
      System.out.println("A2KArea " + version);
      for (int n = 0; n < args.length; n++) {
        s = args[n];
        if (s.length() < 2) continue;
        if (s.charAt(0) == '-') {
          switch (s.charAt(1)) {
            case 'd' :
              d = A2KArray.makeFloat(s.substring(2));
              if (Float.isNaN(d)) throw new Exception();
              if (d <= 0)  throw new Exception();
              break;
            case 'e' :                                             //-2014-06-22
              edge = true;
              break;             
            case 'h':
              help();
            case 'n':
              String[] ss = s.substring(2).split(",");
              if (ss.length != 2) throw new Exception();
              nx = Integer.parseInt(ss[0]);
              ny = Integer.parseInt(ss[1]);
              if (nx <= 0 || ny <= 0)  throw new Exception();
              r = 0;
              break;
            case 'N' :                                             //-2014-06-22
              name = s.substring(2);
              break;
            case 'r' :
              r = A2KArray.makeFloat(s.substring(2));
              if (Float.isNaN(r)) throw new Exception();
              if (r <= 0)  throw new Exception();
              break;
            case 't':
              type = s.substring(2);
              break;
            case 'x' :
              x = A2KArray.makeFloat(s.substring(2));
              if (Float.isNaN(x)) throw new Exception();
              break;
            case 'y' :
              y = A2KArray.makeFloat(s.substring(2));
              if (Float.isNaN(y)) throw new Exception();
              break;
            default :
              throw new Exception();
          }
        } else {
          work = s;
          s = null;
          File f = new File(work);
          if (!f.isDirectory()) {
            System.out.println("*** \"" + work + "\" ist kein Verzeichnis");
            System.exit(2);
          }
        }
      }
    } catch (Exception e) {
      if (s != null)
        System.out.println("*** Fehler bei der Verarbeitung der Option \"" + s
            + "\"");
      System.exit(1);
    }
    //
    try {
      while (true) {
        if (work == null) {
          System.out.println("*** Es ist kein Arbeitsverzeichnis angegeben!");
          System.exit(3);
        }
        A2KArea aa = new A2KArea(work, type, name, edge);
        if (r > 0)
          aa.setGrid(x, y, d, r);
        else
          aa.setGrid(x, y, d, nx, ny);
        msg = aa.makeAList();
        if (msg != null) break;
        msg = aa.makeArray();
        if (msg != null) break;
        String fn = work + "/" + aa.sName +"-d" + (int) d;
        if (r > 0) 
          fn += "r" + (int) r + "x" + (int) x + "y" + (int) y + "-";
        else 
          fn += "x" + (int) x + "n" + nx + "y" + (int) y + "n" + ny + "-";
        msg = aa.save(fn, true);
        break;
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    if (msg != null)
      System.out.println(msg);
    else
      System.out.println("finished");
  }

}
//=================================================================
// history:
//
// 2004-09-04 1.1.3  lj release
// 2004-09-06 1.1.4  lj save corrected
// 2005-02-26 1.2.0  lj optional rectangular evaluation area
// 2008-03-18 1.3.0  lj correction for buildings, odor components
// 2008-04-30 1.3.1  lj output file name corrected
// 2012-09-04 1.3.3  uj check also for locale "de" as alternative to "german"
// 2014-06-22 1.3.4  uj options -e and -N
// 2015-02-21 1.3.5  uj A2KArray: format factor extraction corrected
// 2016-06-03 1.3.6  uj internal adjustments (local nnx, nny, ddd)
//
//=================================================================